"""Shared test fixtures and utilities."""

import pytest


class MockTestItem:
    """Mock test item for testing purposes."""

    def __init__(
        self,
        name: str,
        file_path: str = "test_file.py",
        outcome: str = "passed",
    ):
        """Initialize mock test item with a name and file path."""
        self.name = name
        self.fspath = file_path
        self.nodeid = name
        self.outcome = outcome

    def __str__(self) -> str:
        """Return string representation of the test item."""
        return f"MockTestItem({self.name})"

    def __repr__(self) -> str:
        """Return detailed representation of the test item."""
        return f"MockTestItem({self.name})"

    def __eq__(self, other) -> bool:
        """Check equality based on name and file path."""
        if not isinstance(other, MockTestItem):
            return False
        return self.name == other.name and self.fspath == other.fspath

    def __hash__(self) -> int:
        """Make the object hashable based on name and file path."""
        return hash((self.name, self.fspath))


class MockConfig:
    """A mock config object."""

    def __init__(self, options=None):
        """Initialize the mock config."""
        self.options = options or {}

    def getoption(self, name, default=None):
        """Get an option from the mock config."""
        return self.options.get(name, default)


@pytest.fixture
def mock_test_item():
    """Provide a MockTestItem instance as a fixture."""
    return MockTestItem


@pytest.fixture
def mock_config():
    """Provide a MockConfig instance as a fixture."""
    return MockConfig


@pytest.fixture
def sample_test_items():
    """Provide a list of sample test items for testing."""
    return [
        MockTestItem("gamma"),
        MockTestItem("beta"),
        MockTestItem("delta"),
        MockTestItem("alpha"),
    ]


@pytest.fixture
def multi_file_test_items():
    """Provide test items from multiple files for testing file-based shuffling."""
    return [
        MockTestItem("test_a1", "test_file_a.py"),
        MockTestItem("test_a2", "test_file_a.py"),
        MockTestItem("test_a3", "test_file_a.py"),
        MockTestItem("test_b1", "test_file_b.py"),
        MockTestItem("test_b2", "test_file_b.py"),
        MockTestItem("test_c1", "test_file_c.py"),
        MockTestItem("test_c2", "test_file_c.py"),
        MockTestItem("test_c3", "test_file_c.py"),
        MockTestItem("test_c4", "test_file_c.py"),
    ]



"""Tests for the main plugin functionality."""

# ruff: noqa: PLR2004

from pytest_brightest.plugin import (
    BrightestPlugin,
    _get_brightest_data,
    _plugin,
    pytest_addoption,
    pytest_collection_modifyitems,
    pytest_configure,
    pytest_runtest_logreport,
    pytest_sessionfinish,
)


class TestBrightestPlugin:
    """Test the BrightestPlugin class."""

    def test_configure_disabled(self, mock_config):
        """Test that the plugin is disabled by default."""
        plugin = BrightestPlugin()
        config = mock_config()
        plugin.configure(config)
        assert not plugin.enabled

    def test_configure_enabled(self, mock_config, mocker):
        """Test that the plugin can be enabled."""
        mocker.patch("pytest_brightest.plugin.console.print")
        mocker.patch(
            "pytest_brightest.plugin.setup_json_report_plugin",
            return_value=True,
        )
        plugin = BrightestPlugin()
        config = mock_config({"--brightest": True})
        plugin.configure(config)
        assert plugin.enabled

    def test_configure_shuffle(self, mock_config, mocker):
        """Test that the plugin can be configured to shuffle."""
        mocker.patch("pytest_brightest.plugin.console.print")
        mocker.patch(
            "pytest_brightest.plugin.setup_json_report_plugin",
            return_value=True,
        )
        plugin = BrightestPlugin()
        config = mock_config(
            {
                "--brightest": True,
                "--reorder-by-technique": "shuffle",
                "--seed": 42,
            }
        )
        plugin.configure(config)
        assert plugin.shuffle_enabled
        assert plugin.seed == 42

    def test_configure_reorder(self, mock_config, mocker):
        """Test that the plugin can be configured to reorder."""
        mocker.patch("pytest_brightest.plugin.console.print")
        mocker.patch(
            "pytest_brightest.plugin.setup_json_report_plugin",
            return_value=True,
        )
        plugin = BrightestPlugin()
        config = mock_config(
            {
                "--brightest": True,
                "--reorder-by-technique": "cost",
                "--reorder-in-direction": "ascending",
            }
        )
        plugin.configure(config)
        assert plugin.reorder_enabled
        assert plugin.reorder_by == "cost"
        assert plugin.reorder == "ascending"

    def test_shuffle_tests(self, mock_config, mock_test_item, mocker):
        """Test that the plugin can shuffle tests."""
        mocker.patch("pytest_brightest.plugin.console.print")
        mocker.patch(
            "pytest_brightest.plugin.setup_json_report_plugin",
            return_value=True,
        )
        plugin = BrightestPlugin()
        config = mock_config(
            {
                "--brightest": True,
                "--reorder-by-technique": "shuffle",
                "--seed": 42,
            }
        )
        plugin.configure(config)
        items = [
            mock_test_item("one"),
            mock_test_item("two"),
            mock_test_item("three"),
        ]
        plugin.shuffle_tests(items)
        assert [item.name for item in items] == ["two", "one", "three"]

    def test_reorder_tests(
        self, tmp_path, mock_config, mock_test_item, mocker
    ):
        """Test that the plugin can reorder tests."""
        mocker.patch("pytest_brightest.plugin.console.print")
        mocker.patch(
            "pytest_brightest.plugin.setup_json_report_plugin",
            return_value=True,
        )
        _ = tmp_path
        plugin = BrightestPlugin()
        config = mock_config(
            {
                "--brightest": True,
                "--reorder-by-technique": "cost",
                "--reorder-in-direction": "ascending",
            }
        )
        plugin.configure(config)
        items = [mock_test_item("slow"), mock_test_item("fast")]
        plugin.reorder_tests(items)
        # this test is not complete as it requires a json file
        assert [item.name for item in items] == ["slow", "fast"]

    def test_configure_shuffle_with_direction_warning(
        self, mock_config, mocker
    ):
        """Test that a warning is issued when shuffling with a direction."""
        mock_console_print = mocker.patch(
            "pytest_brightest.plugin.console.print"
        )
        mocker.patch(
            "pytest_brightest.plugin.setup_json_report_plugin",
            return_value=True,
        )
        plugin = BrightestPlugin()
        config = mock_config(
            {
                "--brightest": True,
                "--reorder-by-technique": "shuffle",
                "--reorder-in-direction": "descending",
            }
        )
        plugin.configure(config)
        mock_console_print.assert_any_call(
            ":high_brightness: pytest-brightest: Warning: --reorder-in-direction is ignored when --reorder-by-technique is 'shuffle'"
        )

    def test_configure_json_report_setup_fails(self, mock_config, mocker):
        """Test that a warning is issued when json report setup fails."""
        mock_console_print = mocker.patch(
            "pytest_brightest.plugin.console.print"
        )
        mocker.patch(
            "pytest_brightest.plugin.setup_json_report_plugin",
            return_value=False,
        )
        plugin = BrightestPlugin()
        config = mock_config({"--brightest": True})
        plugin.configure(config)
        mock_console_print.assert_any_call(
            ":high_brightness: pytest-brightest: pytest-json-report setup failed, certain features disabled"
        )


def test_brightestplugin_record_test_failure():
    """Test recording test failures in BrightestPlugin."""
    plugin = BrightestPlugin()
    plugin.record_test_failure("mod1::test1")
    assert plugin.current_session_failures["mod1"] == 1
    plugin.record_test_failure("mod1::test1")
    assert plugin.current_session_failures["mod1"] == 2
    plugin.record_test_failure("")
    # Should not raise


def test_brightestplugin_store_session_items(mock_test_item):
    """Test storing session items in BrightestPlugin."""
    plugin = BrightestPlugin()
    items = [mock_test_item("a"), mock_test_item("b")]
    plugin.store_session_items(items)
    assert plugin.session_items == items


class TestHooks:
    """A container for all the tests of the hooks."""

    def test_pytest_addoption(self, mocker):
        """Test that the command line options are added."""
        parser = mocker.MagicMock()
        parser.getgroup.return_value = mocker.MagicMock()
        pytest_addoption(parser)
        assert parser.getgroup.called
        assert parser.getgroup.return_value.addoption.call_count == 5

    def test_pytest_configure(self, mocker, mock_config):
        """Test that the plugin is configured."""
        mocker.patch.object(_plugin, "configure")
        config = mock_config({"--brightest": True})
        pytest_configure(config)
        _plugin.configure.assert_called_once_with(config)  # type: ignore

    def test_pytest_collection_modifyitems_first_strategy(
        self, mocker, mock_config, mock_test_item
    ):
        """Test that pytest_collection_modifyitems modifies the items."""
        # mock the _plugin instance and its methods
        mock_plugin = mocker.patch(
            "pytest_brightest.plugin._plugin", autospec=True
        )
        mock_plugin.enabled = True
        mock_plugin.reorder_enabled = True
        mock_plugin.shuffle_enabled = False
        mock_plugin.technique = None
        config = mock_config()
        items = [mock_test_item("one"), mock_test_item("two")]
        pytest_collection_modifyitems(config, items)
        mock_plugin.reorder_tests.assert_called_once_with(items)

    def test_pytest_collection_modifyitems_second_strategy(
        self, mocker, mock_config, mock_test_item
    ):
        """Test that pytest_collection_modifyitems modifies the items."""
        # mock the _plugin instance and its methods
        mock_plugin = mocker.patch(
            "pytest_brightest.plugin._plugin", autospec=True
        )
        mock_plugin.enabled = True
        mock_plugin.reorder_enabled = False
        mock_plugin.shuffle_enabled = True
        mock_plugin.technique = None
        config = mock_config()
        items = [mock_test_item("one"), mock_test_item("two")]
        pytest_collection_modifyitems(config, items)
        mock_plugin.shuffle_tests.assert_called_once_with(items)

    def test_pytest_collection_modifyitems_default_to_reorder(
        self, mocker, mock_config, mock_test_item
    ):
        """Test that pytest_collection_modifyitems modifies the items."""
        # mock the _plugin instance and its methods
        mock_plugin = mocker.patch(
            "pytest_brightest.plugin._plugin", autospec=True
        )
        mock_plugin.enabled = True
        mock_plugin.reorder_enabled = True
        mock_plugin.shuffle_enabled = True
        mock_plugin.technique = None
        config = mock_config()
        items = [mock_test_item("one"), mock_test_item("two")]
        pytest_collection_modifyitems(config, items)
        mock_plugin.reorder_tests.assert_called_once_with(items)

    def test_pytest_collection_modifyitems_reorder_and_shuffle_prefers_reorder(
        self, mocker, mock_config, mock_test_item
    ):
        """Test that reordering is preferred when reordering and shuffling are enabled."""
        # mock the _plugin instance and its methods
        mock_plugin = mocker.patch(
            "pytest_brightest.plugin._plugin", autospec=True
        )
        mock_plugin.enabled = True
        mock_plugin.reorder_enabled = True
        mock_plugin.shuffle_enabled = True
        mock_plugin.technique = None
        config = mock_config()
        items = [mock_test_item("one"), mock_test_item("two")]
        pytest_collection_modifyitems(config, items)
        mock_plugin.reorder_tests.assert_called_once_with(items)
        mock_plugin.shuffle_tests.assert_not_called()

    def test_pytest_runtest_logreport(self, mocker):
        """Test that pytest_runtest_logreport records failures."""
        mock_plugin = mocker.patch(
            "pytest_brightest.plugin._plugin", autospec=True
        )
        mock_plugin.enabled = True
        mock_plugin.technique = "failure"
        report = mocker.MagicMock()
        report.failed = True
        report.when = "call"
        report.nodeid = "test_node"
        pytest_runtest_logreport(report)
        mock_plugin.record_test_failure.assert_called_once_with("test_node")

    def test_pytest_sessionfinish_no_json_file(self, mocker, mock_config):
        """Test that pytest_sessionfinish handles no JSON file."""
        _ = mock_config
        mock_plugin = mocker.patch(
            "pytest_brightest.plugin._plugin", autospec=True
        )
        mock_plugin.enabled = True
        mock_plugin.technique = None
        mock_plugin.brightest_json_file = "non_existent.json"
        mocker.patch("pathlib.Path.exists", return_value=False)
        mock_console_print = mocker.patch(
            "pytest_brightest.plugin.console.print"
        )
        pytest_sessionfinish(mocker.MagicMock(), 0)
        assert mock_console_print.call_count == 3
        mock_console_print.assert_any_call(
            ":high_brightness: pytest-brightest: There is no JSON file created by pytest-json-report"
        )

    def test_pytest_sessionfinish_with_json_file(self, mocker, mock_config):
        """Test that pytest_sessionfinish processes JSON file."""
        _ = mock_config
        mock_plugin = mocker.patch(
            "pytest_brightest.plugin._plugin", autospec=True
        )
        mock_plugin.enabled = True
        mock_plugin.brightest_json_file = "existent.json"
        mock_plugin.technique = "cost"
        mock_plugin.focus = "tests-across-modules"
        mock_plugin.direction = "ascending"
        mock_plugin.seed = 123
        mock_path_exists = mocker.patch(
            "pathlib.Path.exists", return_value=True
        )
        mock_json_load = mocker.patch("json.load", return_value={"tests": []})
        mock_json_dump = mocker.patch("json.dump")
        mock_console_print = mocker.patch(
            "pytest_brightest.plugin.console.print"
        )

        mock_file_handle = mocker.MagicMock()
        mocker.patch("pathlib.Path.open", return_value=mock_file_handle)
        mocker.patch(
            "pathlib.Path.stat", return_value=mocker.MagicMock(st_size=100)
        )
        # mock _plugin.reorderer directly
        mock_plugin.reorderer = mocker.MagicMock()
        mock_plugin.reorderer.get_test_total_duration.return_value = (
            0.5  # example return value
        )

        mock_session = mocker.MagicMock()
        mock_session.items = []
        pytest_sessionfinish(mock_session, 0)
        mock_path_exists.assert_called_once_with()
        mock_json_load.assert_called_once_with(
            mock_file_handle.__enter__.return_value
        )
        mock_json_dump.assert_called_once()
        assert mock_console_print.call_count == 3
        mock_console_print.assert_any_call(
            f":flashlight: pytest-brightest: pytest-json-report detected at {mock_plugin.brightest_json_file}"
        )

    def test_pytest_sessionfinish_failure_module_counts(
        self, mocker, mock_config, tmp_path, mock_test_item
    ):
        """Test that pytest_sessionfinish saves module failure counts for failure reordering."""
        _ = mock_config
        mock_plugin = mocker.patch(
            "pytest_brightest.plugin._plugin", autospec=True
        )
        mock_plugin.enabled = True
        mock_plugin.brightest_json_file = str(tmp_path / "report.json")
        mock_plugin.technique = "failure"
        mock_plugin.focus = "modules-within-suite"
        mock_plugin.direction = "descending"
        mock_plugin.current_session_failures = {
            "module_a.py": 1,
            "module_b.py": 2,
            "module_c.py": 0,
        }
        # mock _plugin.reorderer directly
        mock_plugin.reorderer = mocker.MagicMock()
        mock_plugin.reorderer.last_module_failure_counts = {
            "module_a.py": 1,
            "module_b.py": 2,
            "module_c.py": 0,
        }
        mock_plugin.reorderer.get_test_outcome.side_effect = lambda item: {
            "module_a.py::test_a1": "passed",
            "module_a.py::test_a2": "failed",
            "module_b.py::test_b1": "passed",
            "module_b.py::test_b2": "failed",
            "module_b.py::test_b3": "failed",
            "module_c.py::test_c1": "passed",
        }.get(item.nodeid, "passed")
        mocker.patch("pathlib.Path.exists", return_value=True)
        mocker.patch("json.load", return_value={"tests": []})
        mock_json_dump = mocker.patch("json.dump")
        mocker.patch("pytest_brightest.plugin.console.print")

        mock_file_handle = mocker.MagicMock()
        mocker.patch("pathlib.Path.open", return_value=mock_file_handle)
        mocker.patch(
            "pathlib.Path.stat", return_value=mocker.MagicMock(st_size=100)
        )
        mock_session = mocker.MagicMock()
        mock_session.items = [
            mock_test_item("module_a.py::test_a1", outcome="passed"),
            mock_test_item("module_a.py::test_a2", outcome="failed"),
            mock_test_item("module_b.py::test_b1", outcome="passed"),
            mock_test_item("module_b.py::test_b2", outcome="failed"),
            mock_test_item("module_b.py::test_b3", outcome="failed"),
            mock_test_item("module_c.py::test_c1", outcome="passed"),
        ]
        pytest_sessionfinish(mock_session, 0)
        mock_json_dump.assert_called_once()
        args, kwargs = mock_json_dump.call_args
        _ = kwargs
        dumped_data = args[0]
        assert dumped_data["brightest"]["module_failure_counts"] == {
            "module_a.py": 1,
            "module_b.py": 2,
            "module_c.py": 0,
        }


def test_get_brightest_data_all_branches(mocker, mock_test_item):
    """Test _get_brightest_data for all branches."""
    mock_plugin = mocker.patch(
        "pytest_brightest.plugin._plugin", autospec=True
    )
    mock_session = mocker.MagicMock()
    # technique: cost, Focus: modules-within-suite
    mock_plugin.technique = "cost"
    mock_plugin.focus = "modules-within-suite"
    mock_plugin.reorderer = mocker.MagicMock()
    mock_plugin.reorderer.get_test_total_duration.return_value = 1.0
    mock_session.items = [mock_test_item("mod1::test1")]
    data = _get_brightest_data(mock_session)
    assert "current_module_costs" in data
    # technique: cost, Focus: tests-within-module
    mock_plugin.focus = "tests-within-module"
    data = _get_brightest_data(mock_session)
    assert "current_test_costs" in data
    # technique: name, Focus: modules-within-suite
    mock_plugin.technique = "name"
    mock_plugin.focus = "modules-within-suite"
    mock_session.items = [mock_test_item("mod1::test1")]
    data = _get_brightest_data(mock_session)
    assert "current_module_order" in data
    # technique: name, Focus: tests-across-modules
    mock_plugin.focus = "tests-across-modules"
    data = _get_brightest_data(mock_session)
    assert "current_test_order" in data
    # technique: name, Focus: tests-within-module
    mock_plugin.focus = "tests-within-module"
    data = _get_brightest_data(mock_session)
    assert "current_module_tests" in data



"""Test the TestReorderer class."""

import json

import pytest

from pytest_brightest.reorder import (
    ReordererOfTests,
    create_reorderer,
    setup_json_report_plugin,
)


def test_create_reorderer():
    """Test the create_reorderer function."""
    reorderer = create_reorderer()
    assert isinstance(reorderer, ReordererOfTests)
    assert (
        reorderer.json_report_path == ".pytest_cache/pytest-json-report.json"
    )
    reorderer = create_reorderer("custom.json")
    assert isinstance(reorderer, ReordererOfTests)
    assert reorderer.json_report_path == "custom.json"


def test_load_test_data_json_decode_error(tmp_path):
    """Test loading test data with JSON decode error."""
    json_path = tmp_path / "bad.json"
    json_path.write_text("{bad json}")
    reorderer = ReordererOfTests(str(json_path))
    assert not reorderer.has_test_data()


def test_get_prior_data_for_reordering_all_branches(tmp_path, mock_test_item):
    """Test getting prior data for all reordering branches."""
    json_path = tmp_path / "report.json"
    data = {
        "tests": [
            {
                "nodeid": "mod1::t1",
                "call": {"duration": 1.0},
                "outcome": "passed",
            },
            {
                "nodeid": "mod1::t2",
                "call": {"duration": 2.0},
                "outcome": "failed",
            },
            {
                "nodeid": "mod2::t3",
                "call": {"duration": 3.0},
                "outcome": "error",
            },
        ]
    }
    json_path.write_text(json.dumps(data))
    reorderer = ReordererOfTests(str(json_path))
    items = [
        mock_test_item("mod1::t1"),
        mock_test_item("mod1::t2"),
        mock_test_item("mod2::t3"),
    ]
    # COST, MODULES_WITHIN_SUITE
    d = reorderer.get_prior_data_for_reordering(
        items, "cost", "modules-within-suite"
    )
    assert "prior_module_costs" in d
    # COST, TESTS_WITHIN_MODULE
    d = reorderer.get_prior_data_for_reordering(
        items, "cost", "tests-within-module"
    )
    assert "prior_test_costs" in d
    # COST, TESTS_ACROSS_MODULES
    d = reorderer.get_prior_data_for_reordering(
        items, "cost", "tests-across-modules"
    )
    assert "prior_test_costs" in d
    # NAME, MODULES_WITHIN_SUITE
    d = reorderer.get_prior_data_for_reordering(
        items, "name", "modules-within-suite"
    )
    assert "prior_module_order" in d
    # NAME, TESTS_ACROSS_MODULES
    d = reorderer.get_prior_data_for_reordering(
        items, "name", "tests-across-modules"
    )
    assert "prior_test_order" in d
    # NAME, TESTS_WITHIN_MODULE
    d = reorderer.get_prior_data_for_reordering(
        items, "name", "tests-within-module"
    )
    assert "prior_module_tests" in d
    # FAILURE, MODULES_WITHIN_SUITE
    d = reorderer.get_prior_data_for_reordering(
        items, "failure", "modules-within-suite"
    )
    assert "prior_module_failure_counts" in d


def test_reorder_tests_in_place_empty():
    """Test reordering with empty items list."""
    reorderer = ReordererOfTests()
    items = []
    reorderer.reorder_tests_in_place(
        items, "cost", "ascending", "modules-within-suite"
    )
    assert items == []


def test_reorder_tests_in_place_all_branches(tmp_path, mock_test_item, mocker):
    """Test reordering tests in place for all branches."""
    json_path = tmp_path / "report.json"
    data = {
        "tests": [
            {
                "nodeid": "mod1::t1",
                "call": {"duration": 1.0},
                "outcome": "passed",
            },
            {
                "nodeid": "mod1::t2",
                "call": {"duration": 2.0},
                "outcome": "failed",
            },
        ]
    }
    json_path.write_text(json.dumps(data))
    reorderer = ReordererOfTests(str(json_path))
    items = [mock_test_item("mod1::t1"), mock_test_item("mod1::t2")]
    mocker.patch("pytest_brightest.reorder.console.print")
    # modules-within-suite, cost
    reorderer.reorder_tests_in_place(
        items, "cost", "ascending", "modules-within-suite"
    )
    # modules-within-suite, name
    reorderer.reorder_tests_in_place(
        items, "name", "ascending", "modules-within-suite"
    )
    # modules-within-suite, failure
    reorderer.reorder_tests_in_place(
        items, "failure", "ascending", "modules-within-suite"
    )
    # tests-within-module
    reorderer.reorder_tests_in_place(
        items, "cost", "ascending", "tests-within-module"
    )
    # tests-across-modules
    reorderer.reorder_tests_in_place(
        items, "cost", "ascending", "tests-across-modules"
    )

    # modules-within-suite, name
    reorderer.reorder_tests_in_place(
        items, "name", "ascending", "modules-within-suite"
    )
    # modules-within-suite, failure
    reorderer.reorder_tests_in_place(
        items, "failure", "ascending", "modules-within-suite"
    )
    # tests-within-module
    reorderer.reorder_tests_in_place(
        items, "cost", "ascending", "tests-within-module"
    )
    # tests-across-modules
    reorderer.reorder_tests_in_place(
        items, "cost", "ascending", "tests-across-modules"
    )


def test_setup_json_report_plugin_branches(mocker):
    """Test setup_json_report_plugin for all branches and exceptions."""

    class DummyConfig:
        class Option:
            json_report_file = ".report.json"

        option = Option()

        class PluginManager:
            def has_plugin(self, name):
                return name == "pytest_jsonreport"

        pluginmanager = PluginManager()

    config = DummyConfig()
    mocker.patch("pytest_brightest.reorder.console.print")
    assert setup_json_report_plugin(config) is True

    # test ImportError
    def raise_import_error(*args, **kwargs):
        _ = args
        _ = kwargs
        raise ImportError("fail")

    mocker.patch.object(
        config.pluginmanager, "has_plugin", side_effect=raise_import_error
    )
    assert setup_json_report_plugin(config) is False

    # test Exception
    def raise_exception(*args, **kwargs):
        _ = args
        _ = kwargs
        raise Exception("fail")

    mocker.patch.object(
        config.pluginmanager, "has_plugin", side_effect=raise_exception
    )
    assert setup_json_report_plugin(config) is False


def test_load_test_data_key_error(tmp_path):
    """Test loading test data with a KeyError."""
    json_path = tmp_path / "bad.json"
    json_path.write_text('{"tests": [{}]}')  # Missing 'nodeid'
    reorderer = ReordererOfTests(str(json_path))
    assert not reorderer.has_test_data()


class TestReordererOfTests:
    """Test the ReordererOfTests class."""

    def test_load_test_data_no_file(self):
        """Test loading test data when the file does not exist."""
        reorderer = ReordererOfTests("non_existent.json")
        assert not reorderer.has_test_data()

    def test_load_test_data_with_file(self, tmp_path):
        """Test loading test data from a valid JSON file."""
        json_path = tmp_path / "report.json"
        data = {
            "tests": [
                {
                    "nodeid": "test_one",
                    "setup": {"duration": 0.1},
                    "call": {"duration": 0.2},
                    "teardown": {"duration": 0.3},
                    "outcome": "passed",
                }
            ]
        }
        json_path.write_text(json.dumps(data))
        reorderer = ReordererOfTests(str(json_path))
        assert reorderer.has_test_data()
        assert "test_one" in reorderer.test_data
        assert reorderer.test_data["test_one"][
            "total_duration"
        ] == pytest.approx(0.6)
        assert reorderer.test_data["test_one"]["outcome"] == "passed"

    def test_get_test_total_duration(self, mock_test_item):
        """Test getting the total duration of a test."""
        reorderer = ReordererOfTests()
        reorderer.test_data = {
            "test_one": {"total_duration": 1.23, "outcome": "passed"}
        }
        item = mock_test_item("test_one")
        assert reorderer.get_test_total_duration(item) == 1.23  # noqa: PLR2004
        item = mock_test_item("test_two")
        assert reorderer.get_test_total_duration(item) == 0.0

    def test_get_test_outcome(self, mock_test_item):
        """Test getting the outcome of a test."""
        reorderer = ReordererOfTests()
        reorderer.test_data = {
            "test_one": {"total_duration": 1.23, "outcome": "failed"}
        }
        item = mock_test_item("test_one")
        assert reorderer.get_test_outcome(item) == "failed"
        item = mock_test_item("test_two")
        assert reorderer.get_test_outcome(item) == "unknown"

    def test_classify_tests_by_outcome(self, mock_test_item):
        """Test classifying tests by their outcome."""
        reorderer = ReordererOfTests()
        reorderer.test_data = {
            "test_pass": {"total_duration": 1, "outcome": "passed"},
            "test_fail": {"total_duration": 1, "outcome": "failed"},
            "test_error": {"total_duration": 1, "outcome": "error"},
        }
        items = [
            mock_test_item("test_pass"),
            mock_test_item("test_fail"),
            mock_test_item("test_error"),
            mock_test_item("test_unknown"),
        ]
        passing, failing = reorderer.classify_tests_by_outcome(items)
        assert [item.name for item in passing] == ["test_pass", "test_unknown"]
        assert [item.name for item in failing] == ["test_fail", "test_error"]

    def test_sort_tests_by_total_duration(self, mock_test_item):
        """Test sorting tests by their total duration."""
        reorderer = ReordererOfTests()
        reorderer.test_data = {
            "test_slow": {"total_duration": 2.0, "outcome": "passed"},
            "test_fast": {"total_duration": 1.0, "outcome": "passed"},
        }
        items = [mock_test_item("test_slow"), mock_test_item("test_fast")]
        sorted_items = reorderer.sort_tests_by_total_duration(items)
        assert [item.name for item in sorted_items] == [
            "test_fast",
            "test_slow",
        ]
        sorted_items = reorderer.sort_tests_by_total_duration(
            items, ascending=False
        )
        assert [item.name for item in sorted_items] == [
            "test_slow",
            "test_fast",
        ]

    def test_reorder_modules_by_cost(self, mock_test_item, mocker):
        """Test reordering modules by their cumulative cost."""
        reorderer = ReordererOfTests()
        reorderer.test_data = {
            "mod1::test1": {"total_duration": 1.0, "outcome": "passed"},
            "mod1::test2": {"total_duration": 2.0, "outcome": "passed"},
            "mod2::test1": {"total_duration": 4.0, "outcome": "passed"},
        }
        items = [
            mock_test_item("mod1::test1"),
            mock_test_item("mod1::test2"),
            mock_test_item("mod2::test1"),
        ]
        mocker.patch("pytest_brightest.reorder.console.print")
        reorderer.reorder_modules_by_cost(items)
        assert [item.name for item in items] == [
            "mod1::test1",
            "mod1::test2",
            "mod2::test1",
        ]
        reorderer.reorder_modules_by_cost(items, ascending=False)
        assert [item.name for item in items] == [
            "mod2::test1",
            "mod1::test1",
            "mod1::test2",
        ]

    def test_reorder_modules_by_name(self, mock_test_item, mocker):
        """Test reordering modules by their name."""
        reorderer = ReordererOfTests()
        items = [
            mock_test_item("mod_b::test1"),
            mock_test_item("mod_a::test1"),
        ]
        mocker.patch("pytest_brightest.reorder.console.print")
        reorderer.reorder_modules_by_name(items)
        assert [item.name for item in items] == [
            "mod_a::test1",
            "mod_b::test1",
        ]
        reorderer.reorder_modules_by_name(items, ascending=False)
        assert [item.name for item in items] == [
            "mod_b::test1",
            "mod_a::test1",
        ]

    def test_reorder_modules_by_failure(self, mock_test_item, mocker):
        """Test reordering modules by their failure count."""
        reorderer = ReordererOfTests()
        reorderer.test_data = {
            "mod_a::test1": {"total_duration": 1, "outcome": "failed"},
            "mod_b::test1": {"total_duration": 1, "outcome": "passed"},
            "mod_b::test2": {"total_duration": 1, "outcome": "failed"},
            "mod_b::test3": {"total_duration": 1, "outcome": "failed"},
        }
        items = [
            mock_test_item("mod_a::test1"),
            mock_test_item("mod_b::test1"),
            mock_test_item("mod_b::test2"),
            mock_test_item("mod_b::test3"),
        ]
        mocker.patch("pytest_brightest.reorder.console.print")
        reorderer.reorder_modules_by_failure(items)
        assert [item.name for item in items] == [
            "mod_a::test1",
            "mod_b::test1",
            "mod_b::test2",
            "mod_b::test3",
        ]
        reorderer.reorder_modules_by_failure(items, ascending=False)
        assert [item.name for item in items] == [
            "mod_b::test1",
            "mod_b::test2",
            "mod_b::test3",
            "mod_a::test1",
        ]

    def test_reorder_tests_within_module(self, mock_test_item, mocker):
        """Test reordering tests within each module."""
        reorderer = ReordererOfTests()
        reorderer.test_data = {
            "mod1::test_slow": {"total_duration": 2.0, "outcome": "passed"},
            "mod1::test_fast": {"total_duration": 1.0, "outcome": "passed"},
        }
        items = [
            mock_test_item("mod1::test_slow"),
            mock_test_item("mod1::test_fast"),
        ]
        mocker.patch("pytest_brightest.reorder.console.print")
        reorderer.reorder_tests_within_module(items, "cost")
        assert [item.name for item in items] == [
            "mod1::test_fast",
            "mod1::test_slow",
        ]
        reorderer.reorder_tests_within_module(items, "name", ascending=False)
        assert [item.name for item in items] == [
            "mod1::test_slow",
            "mod1::test_fast",
        ]
        reorderer.reorder_tests_within_module(items, "name", ascending=True)
        assert [item.name for item in items] == [
            "mod1::test_fast",
            "mod1::test_slow",
        ]

    def test_reorder_tests_across_modules(self, mock_test_item):
        """Test reordering tests across all modules."""
        reorderer = ReordererOfTests()
        reorderer.test_data = {
            "mod2::test_slow": {"total_duration": 2.0, "outcome": "passed"},
            "mod1::test_fast": {"total_duration": 1.0, "outcome": "passed"},
            "mod1::test_fail": {"total_duration": 1.0, "outcome": "failed"},
        }
        items = [
            mock_test_item("mod2::test_slow"),
            mock_test_item("mod1::test_fast"),
            mock_test_item("mod1::test_fail"),
        ]
        reorderer.reorder_tests_across_modules(items, "cost")
        assert [item.name for item in items] == [
            "mod1::test_fast",
            "mod1::test_fail",
            "mod2::test_slow",
        ]
        reorderer.reorder_tests_across_modules(items, "name")
        assert [item.name for item in items] == [
            "mod1::test_fail",
            "mod1::test_fast",
            "mod2::test_slow",
        ]
        reorderer.reorder_tests_across_modules(items, "name", ascending=False)
        assert [item.name for item in items] == [
            "mod2::test_slow",
            "mod1::test_fast",
            "mod1::test_fail",
        ]
        reorderer.reorder_tests_across_modules(items, "failure")
        assert [item.name for item in items] == [
            "mod2::test_slow",
            "mod1::test_fast",
            "mod1::test_fail",
        ]
        reorderer.reorder_tests_across_modules(
            items, "failure", ascending=False
        )
        assert [item.name for item in items] == [
            "mod1::test_fail",
            "mod2::test_slow",
            "mod1::test_fast",
        ]



"""Main plugin implementation for pytest-brightest."""

import json
from datetime import datetime
from pathlib import Path
from typing import Any, Dict, List, Optional

from _pytest.config import Config  # type: ignore
from _pytest.config.argparsing import Parser  # type: ignore
from _pytest.main import Session  # type: ignore
from _pytest.nodes import Item  # type: ignore
from _pytest.reports import TestReport  # type: ignore
from rich.console import Console

from .constants import (
    ASCENDING,
    BRIGHTEST,
    COST,
    CURRENT_MODULE_COSTS,
    CURRENT_MODULE_FAILURE_COUNTS,
    CURRENT_MODULE_ORDER,
    CURRENT_MODULE_TESTS,
    CURRENT_TEST_COSTS,
    CURRENT_TEST_ORDER,
    DEFAULT_FILE_ENCODING,
    DEFAULT_PYTEST_JSON_REPORT_PATH,
    DESCENDING,
    DIRECTION,
    FAILURE,
    FLASHLIGHT_PREFIX,
    FOCUS,
    HIGH_BRIGHTNESS_PREFIX,
    MODULE_COSTS,
    MODULE_FAILURE_COUNTS,
    MODULE_ORDER,
    MODULE_TESTS,
    MODULES_WITHIN_SUITE,
    NAME,
    NEWLINE,
    NODEID,
    NODEID_SEPARATOR,
    SEED,
    SHUFFLE,
    TECHNIQUE,
    TEST_COSTS,
    TEST_ORDER,
    TESTS_ACROSS_MODULES,
    TESTS_WITHIN_MODULE,
    TIMESTAMP,
)
from .reorder import ReordererOfTests, setup_json_report_plugin
from .shuffler import ShufflerOfTests, generate_random_seed

# create a default console
console = Console()


class BrightestPlugin:
    """Main plugin class that handles pytest integration."""

    def __init__(self) -> None:
        """Initialize the plugin with default settings."""
        self.enabled = False
        self.shuffle_enabled = False
        self.shuffle_by: Optional[str] = None
        self.seed: Optional[int] = None
        self.shuffler: Optional[ShufflerOfTests] = None
        self.details = False
        self.reorder_enabled = False
        self.reorder_by: Optional[str] = None
        self.reorder: Optional[str] = None
        self.reorderer: Optional[ReordererOfTests] = None
        self.brightest_json_file: Optional[str] = None
        self.current_session_failures: Dict[str, int] = {}
        self.session_items: List[Item] = []
        self.technique: Optional[str] = None
        self.focus: Optional[str] = None
        self.direction: Optional[str] = None

    def configure(self, config: Config) -> None:
        """Configure the plugin based on command-line options."""
        # check if the plugin is enabled; if it is not
        # enabled then no further configuration steps are taken
        self.enabled = config.getoption("--brightest", False)
        if not self.enabled:
            return
        # configure the name of the file that will contain the
        # JSON file that contains the pytest-json-report data
        self.brightest_json_file = DEFAULT_PYTEST_JSON_REPORT_PATH
        # always set up JSON reporting when brightest is enabled;
        # this ensures generation of test execution data for future reordering
        json_setup_success = setup_json_report_plugin(config)
        # there was a problem configuring the pytest-json-report plugin,
        # display a diagnostic message and to indicate that certain features
        # will not be available during this run of the test suite
        if not json_setup_success:
            console.print(
                f"{HIGH_BRIGHTNESS_PREFIX} pytest-json-report setup failed, certain features disabled"
            )
        # extract the configuration options for reordering and shuffling
        self.technique = config.getoption("--reorder-by-technique")
        self.focus = config.getoption("--reorder-by-focus")
        self.direction = config.getoption("--reorder-in-direction")
        # if the shuffling technique is chosen, then configure the shuffler
        # and alert the person using the plugin if there is a misconfiguration
        if self.technique == SHUFFLE:
            self.shuffle_enabled = True
            self.shuffle_by = self.focus
            if self.shuffle_by is None:
                self.shuffle_by = TESTS_ACROSS_MODULES
            seed_option = config.getoption("--seed", None)
            if seed_option is not None:
                self.seed = int(seed_option)
            else:
                self.seed = generate_random_seed()
            self.shuffler = ShufflerOfTests(self.seed)
            console.print(
                f"{FLASHLIGHT_PREFIX} Shuffling tests by {self.shuffle_by} with seed {self.seed}"
            )
            if self.direction is not None:
                console.print(
                    f"{HIGH_BRIGHTNESS_PREFIX} Warning: --reorder-in-direction is ignored when --reorder-by-technique is 'shuffle'"
                )
        # if the reordering technique is chosen, then configure the reorderer
        elif self.technique in [NAME, COST, FAILURE]:
            self.reorder_enabled = True
            self.reorder_by = self.technique
            self.reorder = self.direction
            if json_setup_success and self.brightest_json_file:
                self.reorderer = ReordererOfTests(self.brightest_json_file)
            console.print(
                f"{FLASHLIGHT_PREFIX} Reordering tests by {self.reorder_by} in {self.reorder} order with focus {self.focus}"
            )

    def record_test_failure(self, nodeid: str) -> None:
        """Record a test failure for the current session."""
        if nodeid:
            module_path = nodeid.split(NODEID_SEPARATOR)[0]
            if module_path not in self.current_session_failures:
                self.current_session_failures[module_path] = 0
            self.current_session_failures[module_path] += 1

    def shuffle_tests(self, items: List[Item]) -> None:
        """Shuffle test items if shuffling is enabled."""
        # if shuffling is enabled and there are items to shuffle, then
        # shuffle them according to the chosen focus
        if self.shuffle_enabled and self.shuffler and items:
            if self.shuffle_by == TESTS_ACROSS_MODULES:
                self.shuffler.shuffle_items_in_place(items)
            elif self.shuffle_by == TESTS_WITHIN_MODULE:
                self.shuffler.shuffle_items_by_file_in_place(items)
            elif self.shuffle_by == MODULES_WITHIN_SUITE:
                self.shuffler.shuffle_files_in_place(items)

    def reorder_tests(self, items: List[Item]) -> None:
        """Reorder test items if reordering is enabled."""
        # if reordering is enabled and there are items to reorder, then
        # reorder them according to the chosen technique, focus, and direction
        if (
            self.reorder_enabled
            and self.reorderer
            and items
            and self.reorder_by
            and self.reorder
            and self.focus
        ):
            self.reorderer.reorder_tests_in_place(
                items, self.reorder_by, self.reorder, self.focus
            )

    def store_session_items(self, items: List[Item]) -> None:
        """Store the session items for later use in data collection."""
        self.session_items = items.copy()


# create a global plugin instance that can be used by the pytest hooks
_plugin = BrightestPlugin()


def pytest_addoption(parser: Parser) -> None:
    """Add command line options for pytest-brightest."""
    group = parser.getgroup("brightest")
    group.addoption(
        "--brightest",
        action="store_true",
        default=False,
        help="Enable the pytest-brightest plugin",
    )
    group.addoption(
        "--seed",
        type=int,
        default=None,
        help="Set the random seed for test shuffling",
    )
    group.addoption(
        "--reorder-by-technique",
        choices=[SHUFFLE, NAME, COST, FAILURE],
        default=None,
        help="Reorder tests by shuffling, name, cost, or failure",
    )
    group.addoption(
        "--reorder-by-focus",
        choices=[
            MODULES_WITHIN_SUITE,
            TESTS_WITHIN_MODULE,
            TESTS_ACROSS_MODULES,
        ],
        default=TESTS_ACROSS_MODULES,
        help="Reorder modules, tests within modules, or tests across modules",
    )
    group.addoption(
        "--reorder-in-direction",
        choices=[ASCENDING, DESCENDING],
        default=ASCENDING,
        help="Reordered tests in ascending or descending order",
    )


def pytest_configure(config: Config) -> None:
    """Configure the plugin when pytest starts."""
    # configure the plugin using the command-line options
    _plugin.configure(config)


def pytest_collection_modifyitems(config: Config, items: List[Item]) -> None:
    """Modify the collected test items by applying reordering and shuffling."""
    # indicate that config parameter is not used
    _ = config
    # if the plugin is enabled, then apply the reordering in one of the
    # specified techniques according to the command-line options
    if _plugin.enabled:
        # store the original items for later use in data collection
        _plugin.store_session_items(items)
        # notes about how to use the pytest-brightest plugin:
        # (a) the plugin allows for either the shuffling of the test
        # suite or the reordering of the test suite, but a person
        # using the plugin cannot use both techniques at the same time
        # (b) the plugin defaults to using one of the reordering
        # techniques as the default if both techniques are (accidentally)
        # specified on the command line by the person using the plugin
        # --> Use the reordering technique
        if _plugin.reorder_enabled:
            _plugin.reorder_tests(items)
        # --> Use the shuffling technique
        elif _plugin.shuffle_enabled:
            _plugin.shuffle_tests(items)


def pytest_runtest_logreport(report: TestReport) -> None:
    """Capture test failures during the current session."""
    if _plugin.enabled and _plugin.technique == FAILURE and report.failed:
        _plugin.record_test_failure(report.nodeid)


def _get_brightest_data(session: Session) -> Dict[str, Any]:  # noqa: PLR0912, PLR0915
    """Collect brightest data for the JSON report."""
    brightest_data: Dict[str, Any] = {
        TIMESTAMP: datetime.now().isoformat(),
        TECHNIQUE: _plugin.technique,
        FOCUS: _plugin.focus,
        DIRECTION: _plugin.direction,
        SEED: _plugin.seed,
    }
    # add prior data that was used for reordering this session
    if (
        _plugin.reorderer
        and _plugin.session_items
        and _plugin.technique
        and _plugin.focus
    ):
        prior_data = _plugin.reorderer.get_prior_data_for_reordering(
            _plugin.session_items, _plugin.technique, _plugin.focus
        )
        brightest_data.update(prior_data)
    # add current session data
    if _plugin.technique == COST and _plugin.reorderer:
        # reload the test data to get the current session's performance data
        # that was just written by pytest-json-report
        _plugin.reorderer.load_test_data()
        current_module_costs: Dict[str, float] = {}
        current_test_costs: Dict[str, float] = {}
        for item in session.items:
            nodeid = getattr(item, NODEID, "")
            if nodeid:
                cost = _plugin.reorderer.get_test_total_duration(item)
                module_path = nodeid.split("::")[0]
                current_module_costs[module_path] = (
                    current_module_costs.get(module_path, 0.0) + cost
                )
                current_test_costs[nodeid] = cost
        if _plugin.focus == MODULES_WITHIN_SUITE:
            brightest_data[CURRENT_MODULE_COSTS] = current_module_costs
        elif _plugin.focus == TESTS_WITHIN_MODULE:
            brightest_data[CURRENT_MODULE_COSTS] = current_module_costs
            brightest_data[CURRENT_TEST_COSTS] = current_test_costs
        elif _plugin.focus == TESTS_ACROSS_MODULES:
            brightest_data[CURRENT_TEST_COSTS] = current_test_costs
        # maintain legacy keys for backward compatibility
        brightest_data[MODULE_COSTS] = current_module_costs
        brightest_data[TEST_COSTS] = current_test_costs
    elif _plugin.technique == NAME:
        if _plugin.focus == MODULES_WITHIN_SUITE:
            current_module_order = []
            for item in session.items:
                nodeid = getattr(item, NODEID, "")
                if nodeid:
                    module_path = nodeid.split("::")[0]
                    if module_path not in current_module_order:
                        current_module_order.append(module_path)
            brightest_data[CURRENT_MODULE_ORDER] = current_module_order
            # maintain legacy key for backward compatibility
            brightest_data[MODULE_ORDER] = current_module_order
        elif _plugin.focus == TESTS_ACROSS_MODULES:
            current_test_order = [
                getattr(item, NODEID, "") for item in session.items
            ]
            brightest_data[CURRENT_TEST_ORDER] = current_test_order
            # maintain legacy key for backward compatibility
            brightest_data[TEST_ORDER] = current_test_order
        elif _plugin.focus == TESTS_WITHIN_MODULE:
            current_module_tests: Dict[str, List[str]] = {}
            for item in session.items:
                nodeid = getattr(item, NODEID, "")
                if nodeid:
                    module_path = nodeid.split("::")[0]
                    if module_path not in current_module_tests:
                        current_module_tests[module_path] = []
                    current_module_tests[module_path].append(nodeid)
            brightest_data[CURRENT_MODULE_TESTS] = current_module_tests
            # maintain legacy key for backward compatibility
            brightest_data[MODULE_TESTS] = current_module_tests
    elif (
        _plugin.technique == FAILURE and _plugin.focus == MODULES_WITHIN_SUITE
    ):
        # save the current session failure counts for future use
        if _plugin.current_session_failures:
            brightest_data[CURRENT_MODULE_FAILURE_COUNTS] = (
                _plugin.current_session_failures
            )
            # maintain legacy key for backward compatibility
            brightest_data[MODULE_FAILURE_COUNTS] = (
                _plugin.current_session_failures
            )
    return brightest_data


def pytest_sessionfinish(session: Session, exitstatus: int) -> None:
    """Check if JSON file from pytest-json-report exists after test session completes."""
    # indicate that these parameters are not used
    _ = exitstatus
    # if the plugin is enabled and a JSON file is specified, then
    # save the diagnostic data to the JSON file
    if _plugin.enabled and _plugin.brightest_json_file:
        json_file = Path(_plugin.brightest_json_file)
        if json_file.exists():
            with json_file.open("r+", encoding=DEFAULT_FILE_ENCODING) as f:
                data = json.load(f)
                data[BRIGHTEST] = _get_brightest_data(session)
                f.seek(0)
                json.dump(data, f, indent=4)
                f.truncate()
            console.print(NEWLINE)
            console.print(
                f":flashlight: pytest-brightest: pytest-json-report detected at {json_file}"
            )
            console.print(
                f":flashlight: pytest-brightest: pytest-json-report created a JSON file of size: {json_file.stat().st_size} bytes"
            )
        else:
            console.print(NEWLINE)
            console.print(
                ":high_brightness: pytest-brightest: There is no JSON file created by pytest-json-report"
            )
            console.print(
                ":high_brightness: pytest-brightest: Use --json-report from pytest-json-report to create the JSON file"
            )



"""Test reordering functionality based on previous test behavior."""

import json
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple

from rich.console import Console

from .constants import (
    ASCENDING,
    BRIGHTEST,
    CALL,
    CALL_DURATION,
    COST,
    DEFAULT_FILE_ENCODING,
    DEFAULT_PYTEST_JSON_REPORT_PATH,
    DURATION,
    EMPTY_STRING,
    ERROR,
    FAILED,
    FAILURE,
    FLASHLIGHT_PREFIX,
    HIGH_BRIGHTNESS_PREFIX,
    JSON_REPORT_FILE,
    MODULES_WITHIN_SUITE,
    NAME,
    NODEID,
    NODEID_SEPARATOR,
    OUTCOME,
    PRIOR_MODULE_COSTS,
    PRIOR_MODULE_FAILURE_COUNTS,
    PRIOR_MODULE_ORDER,
    PRIOR_MODULE_TESTS,
    PRIOR_TEST_COSTS,
    PRIOR_TEST_ORDER,
    PYTEST_CACHE_DIR,
    PYTEST_JSON_REPORT_PLUGIN_NAME,
    REPORT_JSON,
    SETUP,
    SETUP_DURATION,
    TEARDOWN,
    TEARDOWN_DURATION,
    TESTS,
    TESTS_ACROSS_MODULES,
    TESTS_WITHIN_MODULE,
    TOTAL_DURATION,
    UNKNOWN,
    ZERO_COST,
)

if TYPE_CHECKING:
    from _pytest.nodes import Item  # type: ignore


# create a default console
console = Console()


class ReordererOfTests:
    """Handle test reordering based on previous test performance data."""

    def __init__(self, json_report_path: Optional[str] = None):
        """Initialize the reorderer with optional JSON report path."""
        # the pytest-json-report is a JSON file that contains
        # all of the details about the prior run of a pytest test suite
        self.json_report_path = (
            json_report_path or DEFAULT_PYTEST_JSON_REPORT_PATH
        )
        # the test_data dictionary stores details about the test cases
        # that enables them to be reordered; for instance, it stores
        # information about the cumulative execution time of a test
        self.test_data: Dict[str, Dict[str, Any]] = {}
        self.last_module_failure_counts: Optional[Dict[str, int]] = None
        self.brightest_data: Optional[Dict[str, Any]] = None
        # extract the data from the pytest-json-report that was found
        # and store it in the dictionary called test_data
        self.load_test_data()

    def load_test_data(self) -> None:
        """Load test execution data from the pytest-json-report report file."""
        # create a pathlib Path object for the JSON report file; remember
        # that pytest-brightest does not have its own mechanism for collecting
        # data about test execution as it instead relies on the pytest-json-report
        # plugin to generate a JSON report file that contains the test data
        report_path = Path(self.json_report_path)
        # if the report does not exist in the default location then
        # the data from it cannot be extracted and thus the loading
        # function can return early without doing anything
        if not report_path.exists():
            return
        # attempt to read the JSON file and parse it to extract the data
        try:
            with report_path.open("r", encoding=DEFAULT_FILE_ENCODING) as file:
                # load the JSON data from the file
                data = json.load(file)
                # store the brightest data if it exists for historical information
                self.brightest_data = data.get(BRIGHTEST, {})
                # there is data about test cases
                # that were executed in the list of test information
                if TESTS in data:
                    # iterate through each test in the JSON data that
                    # is its own dictionary in a list of the test dictionaries
                    for test in data[TESTS]:
                        # extract the node ID for the test case
                        node_id = test.get(NODEID, EMPTY_STRING)
                        # if the node ID is not empty then it is okay to extract
                        # the setup, call, and teardown durations for the test case
                        if node_id:
                            setup_duration = test.get(SETUP, {}).get(
                                DURATION, 0.0
                            )
                            call_duration = test.get(CALL, {}).get(
                                DURATION, 0.0
                            )
                            teardown_duration = test.get(TEARDOWN, {}).get(
                                DURATION, 0.0
                            )
                            # calculate the total duration (i.e., the cumulative
                            # execution time for the test case) that will include
                            # the reported costs for these three test stages:
                            # --> setup, call, and teardown
                            total_duration = (
                                setup_duration
                                + call_duration
                                + teardown_duration
                            )
                            # store the test data in the dictionary
                            self.test_data[node_id] = {
                                TOTAL_DURATION: total_duration,
                                OUTCOME: test.get(OUTCOME, UNKNOWN),
                                SETUP_DURATION: setup_duration,
                                CALL_DURATION: call_duration,
                                TEARDOWN_DURATION: teardown_duration,
                            }
        # something went wrong while reading the JSON file
        except (json.JSONDecodeError, KeyError, OSError):
            # if there is an error reading the JSON file, then do not
            # attempt to load any data from it and instead just return
            pass

    def get_test_total_duration(self, item: "Item") -> float:
        """Get the total duration of a test item from previous run(s)."""
        # a pytest item has a nodeid attribute that uniquely identifies it
        node_id = getattr(item, NODEID, EMPTY_STRING)
        # retrieve the test information from the test_data dictionary
        test_info = self.test_data.get(node_id, {})
        # return the total duration of the test, or zero if it is not available
        return test_info.get(TOTAL_DURATION, ZERO_COST)

    def get_test_outcome(self, item: "Item") -> str:
        """Get the outcome of a test item from previous run(s)."""
        # a pytest item has a nodeid attribute that uniquely identifies it
        node_id = getattr(item, NODEID, EMPTY_STRING)
        # retrieve the test information from the test_data dictionary
        test_info = self.test_data.get(node_id, {})
        # return the outcome of the test, or unknown if it is not available
        return test_info.get(OUTCOME, UNKNOWN)

    def classify_tests_by_outcome(
        self, items: List["Item"]
    ) -> Tuple[List["Item"], List["Item"]]:
        """Classify tests into passing and failing based on previous outcomes."""
        passing_tests = []
        failing_tests = []
        # iterate over each item and classify it as passing or failing
        for item in items:
            outcome = self.get_test_outcome(item)
            # the outcome is a string that can be "passed", "failed", or "error"
            if outcome in [FAILED, ERROR]:
                failing_tests.append(item)
            else:
                passing_tests.append(item)
        return passing_tests, failing_tests

    def sort_tests_by_total_duration(
        self, items: List["Item"], ascending: bool = True
    ) -> List["Item"]:
        """Sort tests by total duration in ascending or descending order."""
        # use the sorted function to sort the items by their total duration
        return sorted(
            items, key=self.get_test_total_duration, reverse=not ascending
        )

    def get_prior_data_for_reordering(  # noqa: PLR0912
        self, items: List["Item"], technique: str, focus: str
    ) -> Dict[str, Any]:
        """Get the prior data that was used for reordering during this session."""
        prior_data: Dict[str, Any] = {}
        if technique == COST:
            module_costs: Dict[str, float] = {}
            test_costs: Dict[str, float] = {}
            for item in items:
                nodeid = getattr(item, NODEID, EMPTY_STRING)
                if nodeid:
                    cost = self.get_test_total_duration(item)
                    module_path = nodeid.split(NODEID)[0]
                    module_costs[module_path] = (
                        module_costs.get(module_path, 0.0) + cost
                    )
                    test_costs[nodeid] = cost
            if focus == MODULES_WITHIN_SUITE:
                prior_data[PRIOR_MODULE_COSTS] = module_costs
            elif focus == TESTS_WITHIN_MODULE:
                prior_data[PRIOR_MODULE_COSTS] = module_costs
                prior_data[PRIOR_TEST_COSTS] = test_costs
            elif focus == TESTS_ACROSS_MODULES:
                prior_data[PRIOR_TEST_COSTS] = test_costs
        elif technique == NAME:
            if focus == MODULES_WITHIN_SUITE:
                module_order = []
                for item in items:
                    nodeid = getattr(item, NODEID, "")
                    if nodeid:
                        module_path = nodeid.split("::")[0]
                        if module_path not in module_order:
                            module_order.append(module_path)
                prior_data[PRIOR_MODULE_ORDER] = module_order
            elif focus == TESTS_ACROSS_MODULES:
                prior_data[PRIOR_TEST_ORDER] = [
                    getattr(item, NODEID, "") for item in items
                ]
            elif focus == TESTS_WITHIN_MODULE:
                module_tests: Dict[str, List[str]] = {}
                for item in items:
                    nodeid = getattr(item, NODEID, NODEID_SEPARATOR)
                    if nodeid:
                        module_path = nodeid.split(NODEID_SEPARATOR)[0]
                        if module_path not in module_tests:
                            module_tests[module_path] = []
                        module_tests[module_path].append(nodeid)
                prior_data[PRIOR_MODULE_TESTS] = module_tests
        elif technique == FAILURE:
            module_failure_counts: Dict[str, int] = {}
            for item in items:
                nodeid = getattr(item, NODEID, EMPTY_STRING)
                if nodeid:
                    module_path = nodeid.split(NODEID_SEPARATOR)[0]
                    if module_path not in module_failure_counts:
                        module_failure_counts[module_path] = 0
                    if self.get_test_outcome(item) in [FAILED, ERROR]:
                        module_failure_counts[module_path] += 1
            if focus == MODULES_WITHIN_SUITE:
                prior_data[PRIOR_MODULE_FAILURE_COUNTS] = module_failure_counts
        return prior_data

    def reorder_modules_by_cost(
        self, items: List["Item"], ascending: bool = True
    ) -> None:
        """Reorder test modules by their cumulative cost."""
        module_costs: Dict[str, float] = {}
        module_items: Dict[str, List["Item"]] = {}
        # iterate over each item and calculate the cumulative cost of each module
        for item in items:
            nodeid = getattr(item, NODEID, EMPTY_STRING)
            if nodeid:
                # the module path is the part of the nodeid before the "::"
                module_path = nodeid.split(NODEID_SEPARATOR)[0]
                cost = self.get_test_total_duration(item)
                module_costs[module_path] = (
                    module_costs.get(module_path, ZERO_COST) + cost
                )
                if module_path not in module_items:
                    module_items[module_path] = []
                module_items[module_path].append(item)
        # sort the modules by their cumulative cost
        sorted_modules = sorted(
            module_costs.keys(),
            key=lambda m: module_costs[m],
            reverse=not ascending,
        )
        reordered_items = []
        # if there are sorted modules, then add an extra newline
        # to separate the diagnostic output from what appeared before
        if sorted_modules:
            console.print()
        # iterate over the sorted modules and add their items to the reordered list
        for module in sorted_modules:
            # print out the cost of the module using no more than 5 fixed
            # decimal places for the execution time of the test case
            console.print(
                f"{FLASHLIGHT_PREFIX} Module {module} has cost {module_costs[module]:.5f}"
            )
            reordered_items.extend(module_items[module])
        # replace the original list of items with the reordered list
        items[:] = reordered_items

    def reorder_modules_by_name(
        self, items: List["Item"], ascending: bool = True
    ) -> None:
        """Reorder test modules by their name."""
        module_items: Dict[str, List["Item"]] = {}
        # iterate over each item and group them by module
        for item in items:
            nodeid = getattr(item, NODEID, EMPTY_STRING)
            if nodeid:
                module_path = nodeid.split(NODEID_SEPARATOR)[0]
                if module_path not in module_items:
                    module_items[module_path] = []
                module_items[module_path].append(item)
        # sort the modules by their name
        sorted_modules = sorted(module_items.keys(), reverse=not ascending)
        reordered_items = []
        # if there are sorted modules, then add an extra newline
        # to separate the diagnostic output from what appeared before
        if sorted_modules:
            console.print()
        # iterate over the sorted modules and add their items to the reordered list
        for module in sorted_modules:
            console.print(
                f"{FLASHLIGHT_PREFIX} Module {module} is in the reordered suite"
            )
            reordered_items.extend(module_items[module])
        # replace the original list of items with the reordered list
        items[:] = reordered_items

    def reorder_modules_by_failure(
        self, items: List["Item"], ascending: bool = True
    ) -> None:
        """Reorder test modules by their number of failing tests from previous runs."""
        module_failure_counts: Dict[str, int] = {}
        module_items: Dict[str, List["Item"]] = {}
        # iterate over each item and group them by module
        for item in items:
            nodeid = getattr(item, NODEID, EMPTY_STRING)
            if nodeid:
                module_path = nodeid.split(NODEID_SEPARATOR)[0]
                if module_path not in module_failure_counts:
                    module_failure_counts[module_path] = 0
                    module_items[module_path] = []
                # get the test outcome from previous run data
                if self.get_test_outcome(item) in [FAILED, ERROR]:
                    module_failure_counts[module_path] += 1
                module_items[module_path].append(item)
        # store the failure counts for potential use in reporting
        self.last_module_failure_counts = module_failure_counts
        # sort the modules by their failure count
        sorted_modules = sorted(
            module_failure_counts.keys(),
            key=lambda m: module_failure_counts[m],
            reverse=not ascending,
        )
        reordered_items = []
        # if there are sorted modules, then add an extra newline
        # to separate the diagnostic output from what appeared before
        if sorted_modules:
            console.print()
        # iterate over the sorted modules and add their items to the reordered list
        for module in sorted_modules:
            console.print(
                f"{FLASHLIGHT_PREFIX} Module {module} has {module_failure_counts[module]} failing tests from previous run"
            )
            reordered_items.extend(module_items[module])
        # replace the original list of items with the reordered list
        items[:] = reordered_items

    def reorder_tests_within_module(
        self, items: List["Item"], reorder_by: str, ascending: bool = True
    ) -> None:
        """Reorder tests within each module by the specified technique."""
        module_items: Dict[str, List["Item"]] = {}
        module_order: List[str] = []
        # iterate over each item and group them by module
        for item in items:
            nodeid = getattr(item, NODEID, EMPTY_STRING)
            if nodeid:
                module_path = nodeid.split(NODEID_SEPARATOR)[0]
                if module_path not in module_items:
                    module_items[module_path] = []
                    module_order.append(module_path)
                module_items[module_path].append(item)
        reordered_items = []
        # if there are sorted modules, then add an extra newline
        # to separate the diagnostic output from what appeared before
        if module_order:
            console.print()
        # iterate over the modules and reorder the tests within each module
        for module in module_order:
            console.print(
                f":flashlight: pytest-brightest: Reordering tests in module {module}"
            )
            if reorder_by == COST:
                module_items[module].sort(
                    key=self.get_test_total_duration, reverse=not ascending
                )
            elif reorder_by == NAME:
                # when sorting by name, the direction can be ascending or descending
                # and the key for sorting is the nodeid of the test item
                if ascending:
                    module_items[module].sort(
                        key=lambda item: getattr(item, NODEID, EMPTY_STRING),
                        reverse=False,
                    )
                else:
                    module_items[module].sort(
                        key=lambda item: getattr(item, NODEID, EMPTY_STRING),
                        reverse=True,
                    )
            reordered_items.extend(module_items[module])
        # replace the original list of items with the reordered list
        items[:] = reordered_items

    def reorder_tests_in_place(
        self, items: List["Item"], reorder_by: str, reorder: str, focus: str
    ) -> None:
        """Reorder tests in place based on the specified criteria."""
        # it is not possible to reorder an empty list of items
        if not items:
            return
        # the reorder direction is either ascending or descending
        ascending = reorder == ASCENDING
        # reorder the modules within the suite by their cumulative cost
        if focus == MODULES_WITHIN_SUITE:
            if reorder_by == COST:
                self.reorder_modules_by_cost(items, ascending)
            elif reorder_by == NAME:
                self.reorder_modules_by_name(items, ascending)
            elif reorder_by == FAILURE:
                self.reorder_modules_by_failure(items, ascending)
        # reordering tests within a module
        elif focus == TESTS_WITHIN_MODULE:
            self.reorder_tests_within_module(items, reorder_by, ascending)
        # reorder the tests across all modules in the suite
        elif focus == TESTS_ACROSS_MODULES:
            self.reorder_tests_across_modules(items, reorder_by, ascending)

    def reorder_tests_across_modules(
        self, items: List["Item"], reorder_by: str, ascending: bool = True
    ) -> None:
        """Reorder tests across all modules by the specified technique."""
        if reorder_by == COST:
            items.sort(key=self.get_test_total_duration, reverse=not ascending)
        elif reorder_by == NAME:
            items.sort(
                key=lambda item: getattr(item, NAME, EMPTY_STRING),
                reverse=not ascending,
            )
        elif reorder_by == FAILURE:
            passing_tests, failing_tests = self.classify_tests_by_outcome(
                items
            )
            if ascending:
                items[:] = passing_tests + failing_tests
            else:
                items[:] = failing_tests + passing_tests

    def has_test_data(self) -> bool:
        """Check if test performance data is available."""
        return bool(self.test_data)


def create_reorderer(
    json_report_path: Optional[str] = None,
) -> ReordererOfTests:
    """Create a TestReorderer instance."""
    # create a TestReorderer instance that
    # can be used to reorder based on
    # data about the tests from prior run(s)
    return ReordererOfTests(json_report_path)


def setup_json_report_plugin(config) -> bool:
    """Configure pytest-json-report plugin to generate JSON reports automatically."""
    # attempt to configure the pytest-json-report plugin since the data
    # that it produces is useful for certain reordering tasks like,
    # for instance, reordering tests by cumulative execution time
    try:
        # determine whether or not the pytest-json-report plugin is available
        # and display a suitable diagnostic message before running test suite
        plugin_manager = config.pluginmanager
        if plugin_manager.has_plugin(PYTEST_JSON_REPORT_PLUGIN_NAME):
            console.print(
                f"{FLASHLIGHT_PREFIX} Detected the pytest-json-report plugin"
            )
        else:
            console.print(
                f"{FLASHLIGHT_PREFIX} Did not detect pytest-json-report plugin"
            )
        # configure the directory where the pytest-json-report plugin will
        # store its JSON report file used by pytest-brightest for certain
        # tasks like test reordering according to cumulative execution time
        cache_dir = Path(PYTEST_CACHE_DIR)
        cache_dir.mkdir(exist_ok=True)
        json_report_file = DEFAULT_PYTEST_JSON_REPORT_PATH
        # if a JSON report location was already specified by the user through
        # a command-line argument for the pytest-json-report plugin, alert them
        # to the fact that we are storing it internally to manage history better
        if (
            hasattr(config.option, JSON_REPORT_FILE)
            and config.option.json_report_file == REPORT_JSON
        ):
            console.print(
                f"{FLASHLIGHT_PREFIX} Not using the pytest-json-report in {config.option.json_report_file}"
            )
        # set the JSON report file location for pytest-json-report plugin
        # to be the default location that is used by the pytest-brightest plugin
        config.option.json_report_file = json_report_file
        console.print(
            f"{FLASHLIGHT_PREFIX} Using the pytest-json-report with name {json_report_file}"
        )
        return True
    # was not able to import pytest_jsonreport's plugin which means that it cannot
    # be extract to support certain types of prioritization enabled by pytest-brightest
    except ImportError as e:
        console.print(
            f"{HIGH_BRIGHTNESS_PREFIX} pytest-brightest: pytest-json-report not available: {e}"
        )
        return False
    # some other problem occurred and the pytest-brightest plugin cannot use
    # the pytest-json-report plugin
    except Exception as e:
        console.print(
            f"{HIGH_BRIGHTNESS_PREFIX} pytest-brightest: pytest-json report not setup: {e}"
        )
        return False



"""Test the test shuffling functionality for pytest-brightest."""

from pytest_brightest.shuffler import (
    ShufflerOfTests,
    create_shuffler,
    generate_random_seed,
)


def test_shuffle_files_in_place_noop_extra():
    """Test shuffle_files_in_place with empty list does nothing."""
    shuffler = ShufflerOfTests(seed=42)
    items = []
    shuffler.shuffle_files_in_place(items)
    assert items == []


SEED_123 = 123
SEED_456 = 456
SEED_99 = 99
SEED_42 = 42


def test_shuffler_init_and_repr():
    """Test ShufflerOfTests initialization and __repr__."""
    s = ShufflerOfTests()
    assert s.seed is None
    assert hasattr(s, "_random")
    s2 = ShufflerOfTests(SEED_123)
    assert s2.seed == SEED_123
    assert hasattr(s2, "_random")


def test_shuffle_tests_none_and_empty():
    """Test shuffle_tests with None and empty list."""
    s = ShufflerOfTests()
    assert s.shuffle_tests([]) == []


def test_get_set_seed():
    """Test get_seed and set_seed methods of ShufflerOfTests."""
    s = ShufflerOfTests()
    assert s.get_seed() is None
    s.set_seed(99)
    assert s.get_seed() == 99  # noqa: PLR2004
    s.set_seed(None)
    assert s.get_seed() is None


def test_shuffle_items_in_place_branch():
    """Test shuffle_items_in_place with non-empty list."""
    s = ShufflerOfTests()
    items = [1, 2, 3]
    s.shuffle_items_in_place(items)  # type: ignore
    assert set(items) == {1, 2, 3}


def test_shuffle_items_by_file_in_place_branch():
    """Test shuffle_items_by_file_in_place with non-empty list."""

    class Dummy:
        def __init__(self, name):
            self.name = name
            self.fspath = name

    items = [Dummy("a"), Dummy("b")]
    s = ShufflerOfTests()
    s.shuffle_items_by_file_in_place(items)  # type: ignore
    assert sorted([i.name for i in items]) == ["a", "b"]


def test_shuffle_files_in_place_branch():
    """Test shuffle_files_in_place with non-empty list."""

    class Dummy:
        def __init__(self, name):
            self.name = name
            self.fspath = name

    items = [Dummy("a"), Dummy("b")]
    s = ShufflerOfTests()
    s.shuffle_files_in_place(items)  ## type: ignore
    assert sorted([i.name for i in items]) == ["a", "b"]


def test_create_shuffler_no_seed():
    """Test create_shuffler with no seed."""
    s = create_shuffler()
    assert s.get_seed() is None


def test_shuffler_full_init_and_methods():
    """Test ShufflerOfTests full initialization and all methods."""
    # __init__ with None, with int
    s2 = ShufflerOfTests(SEED_123)
    assert s2.seed == SEED_123
    # shuffle_tests with non-empty
    items = [1, 2, 3]
    shuffled = s2.shuffle_tests(items)  # type: ignore
    assert set(shuffled) == {1, 2, 3}
    # get_seed/set_seed
    assert s2.get_seed() == SEED_123
    s2.set_seed(SEED_456)
    assert s2.get_seed() == SEED_456
    # shuffle_items_in_place with non-empty
    s2.shuffle_items_in_place(items)  # type: ignore
    assert set(items) == {1, 2, 3}

    # shuffle_items_by_file_in_place with fallback
    class Dummy:
        def __init__(self, name):
            self.name = name
            self.path = name

    dummies = [Dummy("a"), Dummy("b")]
    s2.shuffle_items_by_file_in_place(dummies)  # type: ignore
    assert sorted([d.name for d in dummies]) == ["a", "b"]
    # shuffle_files_in_place with fallback
    s2.shuffle_files_in_place(dummies)  # type: ignore
    assert sorted([d.name for d in dummies]) == ["a", "b"]


def test_generate_random_seed_range():
    """Test generate_random_seed returns value in valid range."""
    for _ in range(10):
        seed = generate_random_seed()
        assert 1 <= seed <= 2**31 - 1


def test_shuffle_items_in_place(mock_test_item):
    """Test that the shuffler can shuffle items in place."""
    shuffler = ShufflerOfTests(seed=42)
    items = [
        mock_test_item("one"),
        mock_test_item("two"),
        mock_test_item("three"),
    ]
    shuffler.shuffle_items_in_place(items)
    assert [item.name for item in items] == ["two", "one", "three"]


def test_shuffle_items_by_file_in_place_empty_list():
    """Test that the shuffler can shuffle an empty list by file in place."""
    shuffler = ShufflerOfTests(seed=42)
    items = []
    shuffler.shuffle_items_by_file_in_place(items)
    assert items == []


def test_shuffle_items_by_file_in_place_single_file(mock_test_item):
    """Test that the shuffler can shuffle a single file by file in place."""
    shuffler = ShufflerOfTests(seed=42)
    items = [
        mock_test_item("file1_test1"),
        mock_test_item("file1_test2"),
    ]
    for item in items:
        item.fspath = "/path/to/file1.py"
    shuffler.shuffle_items_by_file_in_place(items)
    assert [item.name for item in items] == ["file1_test2", "file1_test1"]


def test_shuffle_items_by_file_in_place(mock_test_item):
    """Test that the shuffler can shuffle items by file in place."""
    shuffler = ShufflerOfTests(seed=42)
    items = [
        mock_test_item("file1_test1"),
        mock_test_item("file1_test2"),
        mock_test_item("file2_test1"),
        mock_test_item("file2_test2"),
    ]
    # mock the fspath attribute
    for item in items:
        if "file1" in item.name:
            item.fspath = "/path/to/file1.py"
        else:
            item.fspath = "/path/to/file2.py"
    shuffler.shuffle_items_by_file_in_place(items)
    assert [item.name for item in items] == [
        "file1_test2",
        "file1_test1",
        "file2_test2",
        "file2_test1",
    ]


def test_shuffle_files_in_place_empty_list():
    """Test that the shuffler can shuffle an empty list of files in place."""
    shuffler = ShufflerOfTests(seed=42)
    items = []
    shuffler.shuffle_files_in_place(items)
    assert items == []


def test_shuffle_files_in_place_single_file(mock_test_item):
    """Test that the shuffler can shuffle a single file in place."""
    shuffler = ShufflerOfTests(seed=42)
    items = [
        mock_test_item("file1_test1"),
        mock_test_item("file1_test2"),
    ]
    for item in items:
        item.fspath = "/path/to/file1.py"
    shuffler.shuffle_files_in_place(items)
    assert [item.name for item in items] == ["file1_test1", "file1_test2"]


def test_shuffle_files_in_place(mock_test_item):
    """Test that the shuffler can shuffle files in place."""
    shuffler = ShufflerOfTests(seed=42)
    items = [
        mock_test_item("file1_test1"),
        mock_test_item("file1_test2"),
        mock_test_item("file2_test1"),
        mock_test_item("file2_test2"),
    ]
    # mock the fspath attribute
    for item in items:
        if "file1" in item.name:
            item.fspath = "/path/to/file1.py"
        else:
            item.fspath = "/path/to/file2.py"
    shuffler.shuffle_files_in_place(items)
    assert [item.name for item in items] == [
        "file2_test1",
        "file2_test2",
        "file1_test1",
        "file1_test2",
    ]


def test_create_shuffler():
    """Test the create_shuffler function."""
    SEED_42 = 42
    shuffler = create_shuffler(seed=SEED_42)
    assert isinstance(shuffler, ShufflerOfTests)
    assert shuffler.get_seed() == SEED_42


def test_generate_random_seed():
    """Test the generate_random_seed function returns an int in range."""
    seed = generate_random_seed()
    assert isinstance(seed, int)
    assert 1 <= seed <= 2**31 - 1


def test_shuffle_items_by_file_in_place_path_fallback():
    """Test shuffle_items_by_file_in_place fallback to PATH/UNKNOWN."""

    class Dummy:
        def __init__(self, name):
            self.name = name
            self.path = f"/dummy/{name}.py"

    items = [Dummy("a"), Dummy("b")]
    shuffler = ShufflerOfTests(seed=42)
    shuffler.shuffle_items_by_file_in_place(items)  # type: ignore
    assert sorted([item.name for item in items]) == ["a", "b"]


def test_shuffle_files_in_place_path_fallback():
    """Test shuffle_files_in_place fallback to PATH/UNKNOWN."""

    class Dummy:
        def __init__(self, name):
            self.name = name
            self.path = f"/dummy/{name}.py"

    items = [Dummy("a"), Dummy("b")]
    shuffler = ShufflerOfTests(seed=42)
    shuffler.shuffle_files_in_place(items)  # type: ignore
    assert sorted([item.name for item in items]) == ["a", "b"]


def test_shuffle_items_in_place_noop():
    """Test shuffle_items_in_place with empty list does nothing."""
    shuffler = ShufflerOfTests(seed=42)
    items = []
    shuffler.shuffle_items_in_place(items)
    assert items == []


def test_shuffle_items_by_file_in_place_noop():
    """Test shuffle_items_by_file_in_place with empty list does nothing."""
    shuffler = ShufflerOfTests(seed=42)
    items = []
    shuffler.shuffle_items_by_file_in_place(items)
    assert items == []


def test_shuffle_files_in_place_noop():
    """Test shuffle_files_in_place with empty list does nothing."""
    shuffler = ShufflerOfTests(seed=SEED_42)
    items = []
    shuffler.shuffle_files_in_place(items)
    assert items == []



"""Test shuffling functionality for pytest-brightest."""

import random
from typing import TYPE_CHECKING, Dict, List, Optional

from .constants import FSPATH, PATH, UNKNOWN

if TYPE_CHECKING:
    from _pytest.nodes import Item  # type: ignore


class ShufflerOfTests:
    """Handles test shuffling with configurable random seeding."""

    def __init__(self, seed: Optional[int] = None):
        """Initialize the shuffler with an optional random seed."""
        self.seed = seed
        # create a new random number generator to ensure that the shuffling
        # is independent of the global random number generator
        self._random = random.Random(seed)

    def shuffle_tests(self, items: List["Item"]) -> List["Item"]:
        """Shuffle a list of test items using the configured random seed."""
        # it is not possible to shuffle an empty list of items
        if not items:
            return items
        # create a copy of the list of items to avoid modifying the original
        shuffled_items = items.copy()
        # shuffle the list of items in place using the random number generator
        self._random.shuffle(shuffled_items)
        return shuffled_items

    def get_seed(self) -> Optional[int]:
        """Get the current random seed."""
        return self.seed

    def set_seed(self, seed: Optional[int]) -> None:
        """Set a new random seed and reinitialize the random generator."""
        self.seed = seed
        # create a new random number generator to ensure that the shuffling
        # is independent of the global random number generator
        self._random = random.Random(seed)

    def shuffle_items_in_place(self, items: List["Item"]) -> None:
        """Shuffle a list of items in place using the configured random seed."""
        # it is not possible to shuffle an empty list of items
        if items:
            # shuffle the list of items in place using the random number generator
            self._random.shuffle(items)

    def shuffle_items_by_file_in_place(self, items: List["Item"]) -> None:
        """Shuffle test items within each file while preserving file order."""
        # it is not possible to shuffle an empty list of items
        if not items:
            return
        # create a dictionary to group items by their file path
        file_groups: Dict[str, List["Item"]] = {}
        # create a list to maintain the order of the files
        file_order = []
        # iterate over each item and group it by its file path
        for item in items:
            # pytest items have an fspath attribute that contains the path to the file
            # and we can also use the path attribute as a fallback
            file_path = getattr(
                item, FSPATH, str(getattr(item, PATH, UNKNOWN))
            )
            file_path_str = str(file_path)
            # if the file path is not already in the file_groups dictionary,
            # add it and also add it to the file_order list
            if file_path_str not in file_groups:
                file_groups[file_path_str] = []
                file_order.append(file_path_str)
            # add the item to the list of items for the current file path
            file_groups[file_path_str].append(item)
        # clear the original list of items so that it can be repopulated
        items.clear()
        # iterate over the file_order list to preserve the original file order
        for file_path in file_order:
            # get the list of items for the current file path
            file_items = file_groups[file_path]
            # shuffle the list of items for the current file path in place
            self._random.shuffle(file_items)
            # add the shuffled items to the original list of items
            items.extend(file_items)

    def shuffle_files_in_place(self, items: List["Item"]) -> None:
        """Shuffle the order of files while preserving test order within each file."""
        # it is not possible to shuffle an empty list of items
        if not items:
            return
        # create a dictionary to group items by their file path
        file_groups: Dict[str, List["Item"]] = {}
        # create a list to maintain the order of the files
        file_order = []
        # iterate over each item and group it by its file path
        for item in items:
            # pytest items have an fspath attribute that contains the path to the file
            # and we can also use the path attribute as a fallback
            file_path = getattr(
                item, FSPATH, str(getattr(item, PATH, UNKNOWN))
            )
            file_path_str = str(file_path)
            # if the file path is not already in the file_groups dictionary,
            # add it and also add it to the file_order list
            if file_path_str not in file_groups:
                file_groups[file_path_str] = []
                file_order.append(file_path_str)
            # add the item to the list of items for the current file path
            file_groups[file_path_str].append(item)
        # shuffle the order of the files in place
        self._random.shuffle(file_order)
        # clear the original list of items so that it can be repopulated
        items.clear()
        # iterate over the shuffled file_order list
        for file_path in file_order:
            # add the items for the current file path to the original list
            items.extend(file_groups[file_path])


def create_shuffler(seed: Optional[int] = None) -> ShufflerOfTests:
    """Define a factory function to create a TestItemShuffler instance."""
    return ShufflerOfTests(seed)


def generate_random_seed() -> int:
    """Generate a random seed for test shuffling."""
    # the upper bound for the random seed is 2**31 - 1 because it is a
    # common value for the maximum value of a 32-bit signed integer
    return random.randint(1, 2**31 - 1)



"""Define constants used by pytest-brightest."""

# define default values associated with reordering
DEFAULT_PYTEST_JSON_REPORT_PATH = ".pytest_cache/pytest-json-report.json"
PYTEST_JSON_REPORT_PLUGIN_NAME = "pytest_jsonreport"
PYTEST_CACHE_DIR = ".pytest_cache"

# define details about the files
DEFAULT_FILE_ENCODING = "utf-8"

# define defaults about the tests
CALL = "call"
DURATION = "duration"
NODEID = "nodeid"
SETUP = "setup"
TEARDOWN = "teardown"
TESTS = "tests"

# define pytest-specific constants
FSPATH = "fspath"
PATH = "path"
NODEID = "nodeid"
NODEID_SEPARATOR = "::"

# define constants for test outcomes
TOTAL_DURATION = "total_duration"
OUTCOME = "outcome"
UNKNOWN = "unknown"
SETUP_DURATION = "setup_duration"
CALL_DURATION = "call_duration"
TEARDOWN_DURATION = "teardown_duration"
FAILED = "failed"
ERROR = "error"

# define constants for json report
JSON_REPORT_FILE = "json_report_file"
REPORT_JSON = ".report.json"

# define the empty string and starting constants
EMPTY_STRING = ""
NEWLINE = "\n"
ZERO_COST = 0.0
BRIGHTEST = "brightest"
TIMESTAMP = "timestamp"
TECHNIQUE = "technique"
FAILURE = "failure"
FOCUS = "focus"
DIRECTION = "direction"
SEED = "seed"
MODULE_COSTS = "module_costs"
TEST_COSTS = "test_costs"
DATETIME_FORMAT = "%Y-%m-%dT%H:%M:%S.%f"
PYTEST_BRIGHTEST_OUT = "pytest-brightest-output"
PYTEST_BRIGHTEST_OUT_JSON = "pytest-brightest-output.json"

# define constants for reordering techniques
SHUFFLE = "shuffle"
NAME = "name"
COST = "cost"

# define constants for reordering focus
MODULES_WITHIN_SUITE = "modules-within-suite"
TESTS_WITHIN_MODULE = "tests-within-module"
TESTS_ACROSS_MODULES = "tests-across-modules"

# define constants for reordering direction
ASCENDING = "ascending"
DESCENDING = "descending"

# define constants for reporting
MODULE_ORDER = "module_order"
TEST_ORDER = "test_order"
MODULE_TESTS = "module_tests"
MODULE_FAILURE_COUNTS = "module_failure_counts"
PRIOR_MODULE_COSTS = "prior_module_costs"
PRIOR_TEST_COSTS = "prior_test_costs"
PRIOR_MODULE_FAILURE_COUNTS = "prior_module_failure_counts"
PRIOR_MODULE_ORDER = "prior_module_order"
PRIOR_TEST_ORDER = "prior_test_order"
PRIOR_MODULE_TESTS = "prior_module_tests"
CURRENT_MODULE_COSTS = "current_module_costs"
CURRENT_TEST_COSTS = "current_test_costs"
CURRENT_MODULE_FAILURE_COUNTS = "current_module_failure_counts"
CURRENT_MODULE_ORDER = "current_module_order"
CURRENT_TEST_ORDER = "current_test_order"
CURRENT_MODULE_TESTS = "current_module_tests"

# diagnostic message prefixes
FLASHLIGHT_PREFIX = ":flashlight: pytest-brightest:"
HIGH_BRIGHTNESS_PREFIX = ":high_brightness: pytest-brightest:"
